"""Functions to easily build example spectra to test the approach."""

import numpy as np
from eddy.fit_annulus import annulus as annulus_class


def gaussian_ensemble(vrot, vrad=0.0, Tb=40., dV=350., tau=None, rms=0.0,
                      dV_chan=30., N=10, oversample=False, PAmin=-np.pi,
                      PAmax=np.pi, linear_sample=True, plot=True,
                      return_annulus=False):
    """
    Model an ensemble of optically thick or thin Gaussian line profiles.

    Args:
        vrot (float): Rotation velocity at the given radius [m/s].
        vrad (optional[float]): Radial velocity at the given radius [m/s].
        Tb (float): Peak brightness temperature for the lines [K].
        dV (float): Doppler linewidth in [m/s].
        tau (optional[float]): Optical depth. If no optical depth is given,
            assume a purely Gaussian profile.
        rms (optional[float]): RMS of the noise in [K].
        dV_chan (optional[float]): Channel width for the velocity axis [m/s].
        N (int): Number of spectra to generate in the ensemble.
        oversample (bool): If provided, oversample the spectral resolution by
            this amount when building the line profiles to account for
            low-resolution effects.
        PAmin (optional[float]): Minimum polar angle in [rad] to consider.
        PAmax (optional[float]): Maximum polar angle in [rad] to consider.
        linear_sample (optional[bool]): Linearlly sample the position angles or
            draw randomly from the PA range given.
        plot (optional[bool]): If true, plot the ensemble of spectra.
        return_annulus (optional[bool]): If true, return a pre-initialised
            annulus instance.

    Returns (if return_annulus is False):
        spectra (ndarray[float]): [theta.size, velax.size] shaped array of the
            spectra in [K].
        theta (ndarray[float]): Array of polar angles of spectra [rad].
        velax (ndarray[float]): Velocity axis in [m/s].

    Returns (if return_annulus is True):
        annulus (annulus instance): Pre-initialized annnulus instance.
    """

    # Take a velocity axis that is twice the shift due to the rotation and
    # include a random offset so lines are not always at the same velocity.

    velax = 2 * vrot
    velax = np.arange(-velax, velax + 1, dV_chan)
    velax += (2.0 * np.random.random() - 1.0) * np.diff(velax)[0]

    # Generate the theta sampling. Either linearly sample the
    # radian range or randomly sample between the bounds.

    if N < 2:
        raise ValueError("Must have at least two spectra.")
    if PAmin >= PAmax:
        raise ValueError("PAmin must be smaller than PAmax.")
    if linear_sample:
        theta = np.linspace(PAmin, PAmax, N + 1)[1:]
        theta += np.diff(theta)[-1] * np.random.random()
    else:
        theta = (PAmax - PAmin) * np.random.random(N) - PAmin
        theta = np.array(sorted(theta))

    # Build the line profiles.

    v_los = vrot * np.cos(theta) + vrad * np.sin(theta)
    if tau is not None:
        Tex = Tb / (1. - np.exp(-tau))
        spectra = [_thick_line(velax, v, dV, Tex, tau, N=N) for v in v_los]
        spectra = np.array(spectra)
    else:
        spectra = [_gaussian(velax, v, dV, Tb, N=N) for v in v_los]
        spectra = np.array(spectra)
    spectra += rms * np.random.randn(spectra.size).reshape(spectra.shape)

    # Plot the profiles to check they're OK and then return.

    annulus = annulus_class(spectra=spectra, theta=theta, velax=velax)
    if plot:
        annulus.plot_spectra()
    if return_annulus:
        return annulus
    return spectra, theta, velax


def flared_disk_ensemble(radius=1.0, inc=30., mstar=1.0, dist=100., Tb=40.,
                         Tb2=21., dV=350., dV2=110., rms=1., dV_chan=30., N=10,
                         PAmin=-np.pi, PAmax=np.pi, linear_sample=True,
                         aspect_ratio=0.3, flaring_angle=1.25, plot=True,
                         return_ensemble=False):
    """
    Create an ensemble of Gaussian lines with contamination from the far side
    of a flared disk (see Rosenfeld et al. 2013 for examples). Unlike
    `gaussian_ensemble` a rotation velocity is not specified, rather a location
    in the disk with a given inclination and stellar mass.

    - Inputs -

    radius          : The radius of the annulus in (arcseconds).
    inc             : Inclination of the disk in (degrees). Must be non-zero.
    mstar           : Mass of the central star in (Msun).
    dist            : Distance to the central star in (pc).
    Tb              : Peak brightness temperature of the main component in (K).
    Tb2             : Peak brightness temperature of the rear component in (K).
                      For CO this should be the freeze-out temperature of CO,
                      so ~21K. See Pinte et al. (2018a) for a discussion.
    dV              : Doppler width of the main line component in (m/s).
    dV2             : Doppler width of the rear line component in (m/s). The
                      default is the thermal width for Tkin = 21K.
    rms             : RMS of the noise in (K).
    dV_chan         : Channel width for the velocity axis (m/s).
    N               : Number of spectra to generate in the ensemble.
    PAmin           : Minimum polar angle to consider.
    PAmax           : Maximum polar angle to consider.
    linear_sample   : Linearlly sample the position angles or draw randomly.
    aspect_ratio    : The aspect ratio of the disk (z / r) at 1".
    flaring_angle   : The flaring angle of the disk.
    plot            : Plot the ensemble.
    return_ensemble : Return a pre-initialised ensemble instance.

    - Outputs -

    spectra         : [N, velax.size] shaped array of the spectra in (K).
    theta           : Array of polar angles of spectra (radians).
    velax           : Velocity axis in (m/s).
    vrot            : Rotation velocity of the front side in (m/s).

    """

    from scipy.interpolate import griddata

    # Check the inclination.
    if inc <= 0.0:
        raise ValueError("Inclination must be positive.")

    # Define the vertical height functions.

    def zfunc(r):
        return aspect_ratio * np.power(r, flaring_angle)

    # Cyclindrical coordinates of the disk midplane.

    if N < 2:
        raise ValueError("Must have at least two spectra.")
    if PAmin >= PAmax:
        raise ValueError("Must have PA_min < PA_max.")
    if linear_sample:
        tdisk = np.linspace(PAmin, PAmax, N + 1)[1:]
        tdisk += np.diff(tdisk)[-1] * np.random.random()
    else:
        tdisk = 2. * np.pi * np.random.random(N) - np.pi
        tdisk = np.array(sorted(tdisk))
    rdisk = radius * np.ones(N)
    zdisk = zfunc(rdisk)

    # Cartesian coordinates of the disk midplane.

    xdisk = rdisk * np.cos(tdisk)
    ydisk = rdisk * np.sin(tdisk)

    # Find the projected sky coordinates of these locations.

    xsky, ysky = _disk_to_sky(xdisk, ydisk, zdisk, inc)

    # Find where these sky coordinates intersect the rear side of the disk.
    # First make a grid of the cylindrical coordinates of the far side
    # of the disk then find their sky coordinates. We then interpolate this
    # using the sky coordinates of the front side to recover the midplane
    # coordaintes which are interesected. There is probably a way to do
    # this analytically but I haven't worked that out yet...

    xgrid = np.linspace(1.2 * xdisk.min(), 1.2 * xdisk.max(), 50)
    ygrid = np.linspace(5.0 * ydisk.min(), 1.2 * ydisk.max(), 50)
    xgrid, ygrid = np.meshgrid(xgrid, ygrid)
    xgrid, ygrid = xgrid.flatten(), ygrid.flatten()
    zgrid = -zfunc(np.hypot(xgrid, ygrid))
    xproj, yproj = _disk_to_sky(xgrid, ygrid, zgrid, inc)

    # Intersected cylindrical coordinates.

    xback = griddata((xproj, yproj), xgrid, (xsky, ysky))
    yback = griddata((xproj, yproj), ygrid, (xsky, ysky))
    rback, tback = np.hypot(xback, yback), np.arctan2(yback, xback)
    zback = zfunc(rback)

    # Calculate the projected velocities for the line centres.

    vrot = _keplerian_velocity(radius, zfunc(radius), 0.0, inc, dist, mstar)
    vfront = _keplerian_velocity(rdisk, zdisk, tdisk, inc, dist, mstar)
    vback = _keplerian_velocity(rback, zback, tback, inc, dist, mstar)

    print("Rotation velocity: %d m/s." % vrot)

    # Generate the line ensemble, similar to `gaussian_ensemble()`.

    velax = 2 * vrot
    velax = np.arange(-velax, velax + 1, dV_chan)
    velax += (2.0 * np.random.random() - 1.0) * np.diff(velax)[0]

    # Include the emission from the far side of the disk. We take the maximum
    # value to mimic the optically thick front side blocking the rear.

    spectra_f = _gaussian(velax[None, :], vfront[:, None], dV, Tb)
    spectra_b = _gaussian(velax[None, :], vback[:, None], dV2, Tb2)
    spectra = np.max([spectra_f, spectra_b], axis=0)
    spectra += rms * np.random.randn(spectra.size).reshape(spectra.shape)

    # Plot the profiles to check they're OK and then return.

    annulus = annulus_class(spectra=spectra, theta=tdisk, velax=velax)
    if plot:
        annulus.plot_spectra()
    if return_ensemble:
        return annulus
    return spectra, tdisk, velax, vrot


def _gaussian(x, x0, dx, A, N=False):
    """Simple Gaussian function with oversampling."""
    if not N:
        return A * np.exp(-np.power((x - x0) / dx, 2))
    xx = np.linspace(x[0], x[-1], x.size * int(N))
    yy = A * np.exp(-np.power((xx - x0) / dx, 2))
    return np.array([np.average(yy[i*N:(i+1)*N]) for i in range(x.size)])


def _thick_line(x, x0, dx, Tex, tau, N=False):
    """Optically thick line profile."""
    if not N:
        return Tex * (1. - np.exp(-_gaussian(x, x0, dx, tau, N=1)))
    xx = np.linspace(x[0], x[-1], x.size * int(N))
    yy = Tex * (1. - np.exp(-_gaussian(xx, x0, dx, tau, N=False)))
    return np.array([np.average(yy[i*N:(i+1)*N]) for i in range(x.size)])


def _disk_to_sky(xdisk, ydisk, zdisk, inc):
    """Convert disk coordinates to sky coordinates."""
    t = zdisk / np.sin(np.radians(inc)) if abs(inc) > 0.0 else 0.0
    ysky = ydisk - t * np.sin(np.radians(inc))
    return xdisk, ysky * np.cos(np.radians(inc))


def _keplerian_velocity(rdisk, zdisk, tdisk, inc, dist, mstar):
    """Projected Keplerian velocity in [m/s]."""
    import scipy.constants as sc
    vkep = sc.G * mstar * 1.988e30 * np.power(rdisk * dist * sc.au, 2.0)
    vkep /= np.power(np.hypot(rdisk, zdisk) * dist * sc.au, 3.0)
    return np.sqrt(vkep) * np.sin(np.radians(inc)) * np.cos(tdisk)



# -*- coding: utf-8 -*-

import os
import numpy as np
from astropy.io import fits
import scipy.constants as sc
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator, MultipleLocator
import warnings

warnings.filterwarnings("ignore")


class datacube(object):
    """
    A ``datacube`` instance to read in data in a ``FITS`` format. This should
    contain all the functions to interact with a spectral line dataset.

    Args:
        path (str): Path to the datacube to load.
        FOV (Optional[float]): If specified, clip the data down to a
            square field of view with sides of `FOV` [arcsec].
        velocity_range (Optional[list]): A tuple or list of the minimum and
            maximum velocities in [m/s] to clip the data cube to.
        fill (Optional[float]): Replace all ``NaN`` values with this value.
        force_center (Optional[bool]): If ``True`` define the spatial axes such
            that they describe offset from the array center in [arcsec]. This
            is useful if the FITS header does not contain axis information.
    """

    flared_niter = 5
    shadowed_extend = 1.5
    shadowed_oversample = 2.0
    shadowed_method = 'nearest'

    msun = 1.98847e30
    fwhm = 2. * np.sqrt(2 * np.log(2))

    def __init__(self, path, FOV=None, velocity_range=None, fill=None,
                 force_center=False):

        # Read in the data

        self.path = path
        self._read_FITS(path=self.path, fill=fill,
                        force_center=force_center)

        # Clip down the cube.

        if FOV is not None:
            self._clip_cube_spatial(FOV / 2.0, initial_load=True)
        if velocity_range is not None:
            self._clip_cube_velocity(*velocity_range)

    # -- PIXEL DEPROJECTION -- #

    def disk_coords(self, x0=0.0, y0=0.0, inc=0.0, PA=0.0, z0=None, psi=None,
                    r_cavity=0.0, r_taper=None, q_taper=1.0, z_func=None,
                    outframe='cylindrical', shadowed=False, force_side=None,
                    **_):
        r"""
        Get the disk coordinates given certain geometrical parameters and an
        emission surface. The emission surface is most simply described as a
        power law profile,

        .. math::

            z(r) = z_0 \times \left(\frac{r}{1^{\prime\prime}}\right)^{\psi}

        where ``z0`` and ``psi`` can be provided by the user. For the case of
        a non-zero ``z0``, but ``psi=1``, we recover the conical surface
        described in Rosenfeld et al. (2013).

        With the increase in spatial resolution afforded by interferometers
        such as ALMA there are a couple of modifications that can be used to
        provide a better match to the data. For example, an inner cavity can be
        included with the ``r_cavity`` argument which makes the transformation:

        .. math::

            \tilde{r} = {\rm max}(0, r - r_{\rm cavity})

        Note that the inclusion of a cavity will mean that other parameters,
        such as ``z0``, would need to change as the radial axis has effectively
        been shifted.

        To account for the drop in emission surface in the outer disk where the
        gas surface density decreases there are two descriptions. The preferred
        way is to include an exponential taper to the power law profile,

        .. math::

            z_{\rm tapered}(r) = z(r) \times \exp\left( -\left[
            \frac{r}{r_{\rm taper}} \right]^{q_{\rm taper}} \right)

        where both ``r_taper`` and ``q_taper`` values must be set.

        If the emission surface is more complex than the analytical form
        described above, users may provide their own function, ``z_func``,
        which should return the emission height in [arcsec] for a midplane
        radius in [arcsec].

        For certain emission surfaces and high inclination disks, the
        transformation from on-sky coordinates to disk-frame coordinates can be
        hindered by the shadowing of certain regions of the disk. This is a
        particularly big problem if the emission surface is not monotonically
        increasing with radius. For some instances, the default deprojection
        algorithm will fail, and a more robust, albeit slower, algormith is
        needed. This can be turned on with ``shadowed=True``.

        We can also include a warp which is parameterized by,

        .. math::

            z_{\rm warp}(r,\, t) = r \times \tan \left(w_i \times \exp\left(-
            \frac{r^2}{2 w_r^2} \right) \times \sin(t - w_t)\right)

        where ``w_i`` is the inclination in [radians] describing the warp at
        the disk center. The width of the warp is given by ``w_r`` [arcsec] and
        ``w_t`` in [radians] is the angle of nodes (where the warp is zero),
        relative to the position angle of the disk, measured east of north.

        .. WARNING::

            The use of warps is largely untested. Use with caution!
            If you are using a warp, increase the number of iterations
            for the inference through ``self.flared_niter`` (by default at
            5). For high inclinations, also set ``shadowed=True``.

        As it is also possible to determine the rotation direction of the disk
        on the sky, we can encode this information in the sign of the
        inclination. A positive inclination describes a clockwise rotating
        disk, while a negative inclination describes an anti-clockwise rotating
        disk. For 2D disks, i.e., those without an emission surface, this will
        not make a difference, but for 3D disks, this will dictate the
        projection of the surface on the sky.

        Args:
            x0 (Optional[float]): Source right ascension offset [arcsec].
            y0 (Optional[float]): Source declination offset [arcsec].
            inc (Optional[float]): Source inclination [degrees]. A positive
                inclination denotes a disk rotating clockwise on the sky, while
                a negative inclination represents a counter-clockwise rotation.
            PA (Optional[float]): Source position angle [degrees]. Measured
                between north and the red-shifted semi-major axis in an
                easterly direction.
            z0 (Optional[float]): Aspect ratio at 1" for the emission surface.
                To get the far side of the disk, make this number negative.
            psi (Optional[float]): Flaring angle for the emission surface.
            z1 (Optional[float]): Correction term for ``z0``.
            phi (Optional[float]): Flaring angle correction term for the
                emission surface.
            r_cavity (Optional[float]): Outer radius of a cavity. Within this
                region the emission surface is taken to be zero.
            w_i (Optional[float]): Warp inclination in [degrees] at the disk
                center.
            w_r (Optional[float]): Scale radius of the warp in [arcsec].
            w_t (Optional[float]): Angle of nodes of the warp in [degrees].
            z_func (Optional[callable]): A user-defined emission surface
                function that will return ``z`` in [arcsec] for a given ``r``
                in [arcsec]. This will override the analytical form.
            outframe (Optional[str]): Frame of reference for the returned
                coordinates. Either ``'cartesian'`` or ``'cylindrical'``.
            shadowed (Optional[bool]): Whether to use the slower, but more
                robust method for deprojecting pixel values.
            force_side (Optional[str]): Force the emission surface to be a
                particular side, i.e., ``force_side='front'`` will force the
                ``z_func`` argument to be positive, while ``force_side='back'``
                will force ``z_func`` to be negative. Setting
                ``force_side=None`` will allow any limit for the emission
                height.

        Returns:
            array, array, array: Three coordinate arrays with ``(r, phi, z)``,
            in units of [arcsec], [radians], [arcsec], if
            ``frame='cylindrical'`` or ``(x, y, z)``, all in units of [arcsec]
            if ``frame='cartesian'``.
        """

        # Check the input variables.

        outframe = outframe.lower()
        if outframe not in ['cylindrical', 'cartesian']:
            raise ValueError("frame must be 'cylindrical' or 'cartesian'.")

        # Apply the inclination convention to be consistent with orbits.

        inc = inc if inc < 90.0 else inc - 180.0

        # Cycle through the different options for pixel deprojection.

        if z0 is None:
            r, t = self._get_midplane_polar_coords(x0, y0, inc, PA)
            z = np.zeros(r.shape)
        elif psi is None:
            r, t, z = self._get_conical_polar_coords(x0, y0, inc, PA, z0)
        else:
            if z_func is None:
                r_taper = np.inf if r_taper is None else r_taper
                def z_func(r_in):
                    r = np.clip(r_in - r_cavity, a_min=0.0, a_max=None)
                    return z0 * r**psi * np.exp(-np.power(r/r_taper, q_taper))
            if shadowed:
                r, t, z = self._get_shadowed_coords(x0, y0, inc, PA, z_func)
            else:
                r, t, z = self._get_flared_coords(x0, y0, inc, PA, z_func)

        # Return the values.

        if outframe == 'cylindrical':
            return r, t, z
        return r * np.cos(t), r * np.sin(t), z

    def disk_to_sky(self, coords, inc, PA, x0=0.0, y0=0.0, frame='cartesian'):
        """
        Project disk-frame coordinates onto the sky plane.

        Args:
            coords (tuple): A tuple of the disk-frame coordinates to transform.
                Must be either cartestian, cylindrical or spherical frames,
                specified by the ``frame`` argument. If only two coordinates
                are given, the input is assumed to be 2D. All spatial
                coordinates should be given in [arcsec], while all angular
                coordinates should be given in [radians].
            inc (float): Inclination of the disk in [deg].
            PA (float): Position angle of the disk, measured Eastwards to the
                red-shifted major axis from North in [deg].
            x0 (Optional[float]): Source right ascension offset in [arcsec].
            y0 (Optional[float]): Source declination offset in [arcsec].
            frame (Optional[str]): Coordinate frame of the disk coordinates,
                either ``'cartesian'``, ``'cylindrical'`` or ``'spherical'``.

        Returns:
            Two arrays representing the projection of the input coordinates
            onto the sky, ``x_sky`` and ``y_sky``.
        """
        try:
            c1, c2, c3 = coords
        except ValueError:
            c1, c2 = coords
            c3 = np.zeros(c1.size)
        if frame.lower() == 'cartesian':
            x, y, z = c1, c2, c3
        elif frame.lower() == 'cylindrical':
            x = c1 * np.cos(c2)
            y = c1 * np.sin(c2)
            z = c3
        elif frame.lower() == 'spherical':
            x = c1 * np.cos(c2) * np.sin(c3)
            y = c1 * np.sin(c2) * np.sin(c3)
            z = c1 * np.cos(c3)
        else:
            raise ValueError("frame_in must be 'cartestian'," +
                             " 'cylindrical' or 'spherical'.")
        inc = np.radians(inc)
        PA = -np.radians(PA + 90.0)
        y_roll = np.cos(inc) * y - np.sin(inc) * z
        x_sky = np.cos(PA) * x - np.sin(PA) * y_roll
        y_sky = np.sin(PA) * x + np.cos(PA) * y_roll
        return x_sky, y_sky

    @staticmethod
    def _rotate_coords(x, y, PA):
        """Rotate (x, y) by PA [deg]."""
        x_rot = y * np.cos(np.radians(PA)) + x * np.sin(np.radians(PA))
        y_rot = x * np.cos(np.radians(PA)) - y * np.sin(np.radians(PA))
        return x_rot, y_rot

    @staticmethod
    def _deproject_coords(x, y, inc):
        """Deproject (x, y) by inc [deg]."""
        return x, y / np.cos(np.radians(inc))

    def _get_cart_sky_coords(self, x0, y0):
        """Return cartesian sky coordinates in [arcsec, arcsec]."""
        return np.meshgrid(self.xaxis - x0, self.yaxis - y0)

    def _get_midplane_cart_coords(self, x0, y0, inc, PA):
        """Return cartesian coordaintes of midplane in [arcsec, arcsec]."""
        x_sky, y_sky = self._get_cart_sky_coords(x0, y0)
        x_rot, y_rot = self._rotate_coords(x_sky, y_sky, PA)
        return datacube._deproject_coords(x_rot, y_rot, inc)

    def _get_midplane_polar_coords(self, x0, y0, inc, PA):
        """Return the polar coordinates of midplane in [arcsec, radians]."""
        x_mid, y_mid = self._get_midplane_cart_coords(x0, y0, inc, PA)
        return np.hypot(y_mid, x_mid), np.arctan2(y_mid, x_mid)

    def _get_conical_cart_coords(self, x0, y0, inc, PA, z0):
        """Return the cartesian coords of a conical surface."""
        inc = np.radians(inc)
        PA = np.radians(PA - 90.0)
        x_sky, y_sky = self._get_cart_sky_coords(x0, y0)
        x_rot = x_sky * np.cos(PA) - y_sky * np.sin(PA)
        y_rot = x_sky * np.sin(PA) + y_sky * np.cos(PA)
        psi = np.tan(z0)
        a = np.cos(2 * inc) + np.cos(2 * psi)
        b = -4.0 * np.sin(psi)**2 * y_rot * np.tan(inc)
        c = -2.0 * np.sin(psi)**2 * (x_rot**2 + y_rot**2 / np.cos(inc)**2)
        t = -b + np.sqrt(b**2 - 4 * a * c) / 2 / a
        x_d = x_rot
        y_d = y_rot / np.cos(inc) + t * np.sin(inc)
        z_d = z0 * np.hypot(x_d, y_d)
        return x_d, y_d, z_d

    def _get_conical_polar_coords(self, x0, y0, inc, PA, z0):
        """Return the cylindrical coords of a conical surface."""
        x_d, y_d, z_d = self._get_conical_cart_coords(x0, y0, inc, PA, z0)
        return np.hypot(y_d, x_d), np.arctan2(y_d, x_d), z_d

    def _get_flared_coords(self, x0, y0, inc, PA, z_func, w_func=None):
        """Return cyclindrical coords of surface in [arcsec, rad, arcsec]."""
        x_mid, y_mid = self._get_midplane_cart_coords(x0, y0, inc, PA)
        r_tmp, t_tmp = np.hypot(x_mid, y_mid), np.arctan2(y_mid, x_mid)
        for _ in range(self.flared_niter):
            y_tmp = y_mid + z_func(r_tmp) * np.tan(np.radians(inc))
            r_tmp = np.hypot(y_tmp, x_mid)
            t_tmp = np.arctan2(y_tmp, x_mid)
        return r_tmp, t_tmp, z_func(r_tmp)

    def _get_shadowed_coords(self, x0, y0, inc, PA, z_func, w_func=None):
        """
        Return cyclindrical coords of surface in [arcsec, rad, arcsec].
        """

        # Make the disk-frame coordinates.

        xdisk, ydisk, rdisk, tdisk = self._get_diskframe_coords()
        zdisk = z_func(rdisk)
        if w_func is not None:
            zdisk += w_func(rdisk, tdisk)

        # Incline the disk.

        inc = np.radians(inc)
        x_dep = xdisk
        y_dep = ydisk * np.cos(inc) - zdisk * np.sin(inc)

        # Remove shadowed pixels.
        # TODO: Check how this handles the bottom side of the disk.

        if inc < 0.0:
            y_dep = np.maximum.accumulate(y_dep, axis=0)
        else:
            y_dep = np.minimum.accumulate(y_dep[::-1], axis=0)[::-1]

        # Rotate and recenter the disk.

        x_rot, y_rot = self._rotate_coords(x_dep, y_dep, PA)
        x_rot, y_rot = x_rot + x0, y_rot + y0

        # Grid the disk.

        from scipy.interpolate import griddata
        disk = (x_rot.flatten(), y_rot.flatten())
        grid = (self.xaxis[None, :], self.yaxis[:, None])
        r_obs = griddata(disk, rdisk.flatten(), grid,
                         method=self.shadowed_method)
        t_obs = griddata(disk, tdisk.flatten(), grid,
                         method=self.shadowed_method)
        return r_obs, t_obs, z_func(r_obs)

    def _get_diskframe_coords(self):
        """Disk-frame coordinates based on the cube axes."""
        x_disk = np.linspace(self.shadowed_extend * self.xaxis[0],
                             self.shadowed_extend * self.xaxis[-1],
                             int(self.nxpix * self.shadowed_oversample))[::-1]
        y_disk = np.linspace(self.shadowed_extend * self.yaxis[0],
                             self.shadowed_extend * self.yaxis[-1],
                             int(self.nypix * self.shadowed_oversample))
        x_disk, y_disk = np.meshgrid(x_disk, y_disk)
        r_disk = np.hypot(x_disk, y_disk)
        t_disk = np.arctan2(y_disk, x_disk)
        return x_disk, y_disk, r_disk, t_disk

    # -- MASKING FUNCTIONS -- #

    def get_mask(self, r_min=None, r_max=None, exclude_r=False, phi_min=None,
                 phi_max=None, exclude_phi=False, abs_phi=False, x0=0.0,
                 y0=0.0, inc=0.0, PA=0.0, z0=0.0, psi=1.0, r_cavity=0.0,
                 r_taper=np.inf, q_taper=1.0, w_i=None, w_r=None, w_t=None,
                 z_func=None, shadowed=False, mask_frame='disk',
                 user_mask=None):
        """
        Returns a 2D mask for pixels in the given region. The mask can be
        specified in either disk-centric coordinates, ``mask_frame='disk'``,
        or on the sky, ``mask_frame='sky'``. If sky-frame coordinates are
        requested, the geometrical parameters (``inc``, ``PA``, ``z0``, etc.)
        are ignored, however the source offsets, ``x0``, ``y0``, are still
        considered.

        Args:
            r_min (Optional[float]): Minimum midplane radius of the annulus in
                [arcsec]. Defaults to minimum deprojected radius.
            r_max (Optional[float]): Maximum midplane radius of the annulus in
                [arcsec]. Defaults to the maximum deprojected radius.
            exclude_r (Optional[bool]): If ``True``, exclude the provided
                radial range rather than include.
            phi_min (Optional[float]): Minimum polar angle of the segment of
                the annulus in [deg]. Note this is the polar angle, not the
                position angle.
            phi_max (Optional[float]): Maximum polar angle of the segment of
                the annulus in [deg]. Note this is the polar angle, not the
                position angle.
            exclude_phi (Optional[bool]): If ``True``, exclude the provided
                polar angle range rather than include it.
            abs_phi (Optional[bool]): If ``True``, take the absolute value of
                the polar angle such that it runs from 0 [deg] to 180 [deg].
            x0 (Optional[float]): Source center offset along the x-axis in
                [arcsec].
            y0 (Optional[float]): Source center offset along the y-axis in
                [arcsec].
            inc (Optional[float]): Inclination of the disk in [degrees].
            PA (Optional[float]): Position angle of the disk in [degrees],
                measured east-of-north towards the redshifted major axis.
            z0 (Optional[float]): Emission height in [arcsec] at a radius of
                1".
            psi (Optional[float]): Flaring angle of the emission surface.
            z_func (Optional[function]): A function which provides
                :math:`z(r)`. Note that no checking will occur to make sure
                this is a valid function.
            shadowed (Optional[bool]): Whether to use the slower, but more
                robust, deprojection method for shadowed disks.

        Returns:
            A 2D array mask matching the shape of a channel.
        """

        # Check the requested frame.

        mask_frame = mask_frame.lower()
        if mask_frame not in ['disk', 'sky']:
            raise ValueError("mask_frame must be 'disk' or 'sky'.")
        if mask_frame == 'sky':
            inc = 0.0
            PA = 0.0

        # Calculate the deprojected pixel coordaintes.

        rvals, pvals = self.disk_coords(x0=x0, y0=y0, inc=inc, PA=PA, z0=z0,
                                        psi=psi, r_cavity=r_cavity,
                                        r_taper=r_taper, q_taper=q_taper,
                                        w_i=w_i, w_r=w_r, w_t=w_t,
                                        z_func=z_func, frame='cylindrical',
                                        shadowed=shadowed)[:2]
        pvals = abs(pvals) if abs_phi else pvals

        # Radial mask.

        r_min = np.nanmin(rvals) if r_min is None else r_min
        r_max = np.nanmax(rvals) if r_max is None else r_max
        if r_min >= r_max:
            raise ValueError("`r_min` must be smaller than `r_max`.")
        r_mask = np.logical_and(rvals >= r_min, rvals <= r_max)
        r_mask = ~r_mask if exclude_r else r_mask

        # Azimuthal mask.

        phi_min = np.nanmin(pvals) if phi_min is None else np.radians(phi_min)
        phi_max = np.nanmax(pvals) if phi_max is None else np.radians(phi_max)
        if phi_min >= phi_max:
            raise ValueError("`PA_min` must be smaller than `PA_max`.")
        phi_mask = np.logical_and(pvals >= phi_min, pvals <= phi_max)
        phi_mask = ~phi_mask if exclude_phi else phi_mask

        # Combine and return.

        mask = r_mask * phi_mask
        if np.sum(mask) == 0:
            raise ValueError("There are zero pixels in the mask.")
        if user_mask is not None:
            mask *= user_mask
        return mask

    # -- DATA I/O -- #

    def _read_FITS(self, path, fill=None, force_center=False):
        """Reads the data from the FITS file."""

        # File names.

        self.path = os.path.expanduser(path)
        self.fname = self.path.split('/')[-1]

        # Read in the data and, if necessary, fill the NaNs with default
        # values. Note that in the case of multiple data fields, we need to
        # think of something different.

        self.header = fits.getheader(path)
        self.data = np.squeeze(fits.getdata(self.path))
        if fill is not None:
            self.data = np.where(np.isfinite(self.data), self.data, fill)

        # Position axes. Two options here, either try to build the axis based
        # on the information in the header, or if force_center=True then return
        # an axis where the offset is relative to the image center

        if force_center:
            self.xaxis = self._forcepositionaxis(a=1)
            self.yaxis = self._forcepositionaxis(a=2)
        else:
            self.xaxis = self._readpositionaxis(a=1)
            self.yaxis = self._readpositionaxis(a=2)
            self.xaxis -= 0.5*self.dpix
            self.yaxis -= 0.5*self.dpix

        # Spectral axis.

        self.nu0 = self._readrestfreq()
        try:
            self.velax = self._readvelocityaxis()
            if self.velax.size > 1:
                self.chan = np.mean(np.diff(self.velax))
            else:
                self.chan = np.nan
            self.freqax = self._readfrequencyaxis()
            if self.chan < 0.0:
                self.data = self.data[::-1]
                self.velax = self.velax[::-1]
                self.freqax = self.freqax[::-1]
                self.chan *= -1.0
        except KeyError:
            self.velax = None
            self.chan = None
            self.freqax = None

        # Check that the data is saved such that increasing indices in x are
        # decreasing in offset counter to the yaxis.

        if np.diff(self.xaxis).mean() > 0.0:
            self.xaxis = self.xaxis[::-1]
            self.data = self.data[:, ::-1]

        # Read the beam properties.

        self._read_beam()

    def _read_beam(self):
        """Reads the beam properties from the header."""
        try:
            if self.header.get('CASAMBM', False):
                beam = fits.open(self.path)[1].data
                beam = np.median([b[:3] for b in beam.view()], axis=0)
                self.bmaj, self.bmin, self.bpa = beam
            else:
                self.bmaj = self.header['bmaj'] * 3600.
                self.bmin = self.header['bmin'] * 3600.
                self.bpa = self.header['bpa']
            self.beamarea_arcsec = self._calculate_beam_area_arcsec()
            self.beamarea_str = self._calculate_beam_area_str()
        except Exception:
            print("WARNING: No beam values found. Assuming pixel as beam.")
            self.bmaj = self.dpix
            self.bmin = self.dpix
            self.bpa = 0.0
            self.beamarea_arcsec = self.dpix**2.0
            self.beamarea_str = np.radians(self.dpix / 3600.)**2.0
        self.bpa %= 180.0

    def print_beam(self):
        """Print the beam properties."""
        print('{:.2f}" x {:.2f}" at {:.1f} deg'.format(*self.beam))

    @property
    def beam(self):
        """Returns beam properties."""
        return self.bmaj, self.bmin, self.bpa

    @property
    def beams_per_pix(self):
        """Number of beams per pixel."""
        return self.dpix**2.0 / self.beamarea_arcsec

    @property
    def pix_per_beam(self):
        """Number of pixels in a beam."""
        return self.beamarea_arcsec / self.dpix**2.0

    def _clip_cube_velocity(self, v_min=None, v_max=None):
        """Clip the cube to within ``vmin`` and ``vmax``."""
        if self.data.ndim == 2:
            raise ValueError("Attaced cube has no velocity axis.")
        v_min = self.velax[0] if v_min is None else v_min
        v_max = self.velax[-1] if v_max is None else v_max
        i = abs(self.velax - v_min).argmin()
        i += 1 if self.velax[i] < v_min else 0
        j = abs(self.velax - v_max).argmin()
        j -= 1 if self.velax[j] > v_max else 0
        self.velax = self.velax[i:j+1]
        self.data = self.data[i:j+1]

    def clip_cube_spatial(self, FOV):
        """Applies the new FOV."""
        if FOV > (self.xaxis.max() - self.xaxis.min()):
            raise ValueError("Cannot apply a larger FOV.")
        xa, xb, ya, yb = self._clip_cube_spatial(FOV / 2.0, False, True)
        self.data = self.data[ya:yb, xa:xb]
        self.error = self.error[ya:yb, xa:xb]
        self.mask = self.mask[ya:yb, xa:xb]
        self.xaxis = self.xaxis[xa:xb]
        self.yaxis = self.yaxis[ya:yb]

    def _clip_cube_spatial(self, radius, initial_load=True, indices=False):
        """Clip the cube plus or minus clip arcseconds from the origin."""
        if radius > min(self.xaxis.max(), self.yaxis.max()):
            print('WARNING: FOV = {:.1f}" larger than '.format(radius * 2)
                  + 'FOV of cube: {:.1f}".'.format(self.xaxis.max() * 2))
        else:
            if initial_load:
                self._original_shape = self.data.shape
            xa = abs(self.xaxis - radius).argmin()
            if self.xaxis[xa] < radius:
                xa -= 1
            xb = abs(self.xaxis + radius).argmin()
            if -self.xaxis[xb] < radius:
                xb += 1
            xb += 1
            ya = abs(self.yaxis + radius).argmin()
            if -self.yaxis[ya] < radius:
                ya -= 1
            yb = abs(self.yaxis - radius).argmin()
            if self.yaxis[yb] < radius:
                yb += 1
            yb += 1
            if initial_load:
                self._xa = xa
                self._xb = xb
                self._ya = ya
                self._yb = yb
            if indices:
                return xa, xb, ya, yb
            if self.data.ndim == 3:
                self.data = self.data[:, ya:yb, xa:xb]
            else:
                self.data = self.data[ya:yb, xa:xb]
            self.xaxis = self.xaxis[xa:xb]
            self.yaxis = self.yaxis[ya:yb]

    @property
    def nxpix(self):
        return self.xaxis.size

    @property
    def nypix(self):
        return self.yaxis.size

    @property
    def dpix(self):
        return np.diff(self.yaxis).mean()

    @property
    def nchan(self):
        if self.velax is not None:
            return self.velax.size
        return 0

    def _readspectralaxis(self, a):
        """Returns the spectral axis in [Hz] or [m/s]."""
        a_len = self.header['naxis%d' % a]
        a_del = self.header['cdelt%d' % a]
        a_pix = self.header['crpix%d' % a]
        a_ref = self.header['crval%d' % a]
        return a_ref + (np.arange(a_len) - a_pix + 1.0) * a_del

    def _readpositionaxis(self, a=1):
        """Returns the position axis in [arcseconds]."""
        if a not in [1, 2]:
            raise ValueError("'a' must be in [1, 2].")
        try:
            a_len = self.header['naxis%d' % a]
            a_del = self.header['cdelt%d' % a]
            a_pix = self.header['crpix%d' % a]
        except KeyError:
            if self._user_pixel_scale is None:
                print('WARNING: No axis information found.')
                _input = input("\t Enter pixel scale size in [arcsec]: ")
                self._user_pixel_scale = float(_input) / 3600.0
            a_len = self.data.shape[-1] if a == 1 else self.data.shape[-2]
            if a == 1:
                a_del = -1.0 * self._user_pixel_scale
            else:
                a_del = 1.0 * self._user_pixel_scale
            a_pix = a_len / 2.0 + 0.5
        axis = (np.arange(a_len) - a_pix + 1.0) * a_del
        return 3600.0 * axis

    def _forcepositionaxis(self, a=1):
        """Returns the axis in [arcsec] assuming image is centered."""
        if a not in [1, 2]:
            raise ValueError("'a' must be in [1, 2].")
        a_len = self.data.shape[-1] if a == 1 else self.data.shape[-2]
        axis = np.arange(a_len).astype('float') - a_len / 2.0
        axis += 0.5 * (abs(axis[0]) - abs(axis[-1]))
        return axis

    def _readrestfreq(self):
        """Read the rest frequency."""
        try:
            nu = self.header['restfreq']
        except KeyError:
            try:
                nu = self.header['restfrq']
            except KeyError:
                try:
                    nu = self.header['crval3']
                except KeyError:
                    nu = np.nan
        return nu

    def _readvelocityaxis(self):
        """Wrapper for _velocityaxis and _spectralaxis."""
        a = 4 if 'stokes' in self.header['ctype3'].lower() else 3
        if 'freq' in self.header['ctype%d' % a].lower():
            specax = self._readspectralaxis(a)
            velax = (self.nu0 - specax) * sc.c
            velax /= self.nu0
        else:
            velax = self._readspectralaxis(a)
        return velax

    def _readfrequencyaxis(self):
        """Returns the frequency axis in [Hz]."""
        a = 4 if 'stokes' in self.header['ctype3'].lower() else 3
        if 'freq' in self.header['ctype3'].lower():
            return self._readspectralaxis(a)
        return self._readrestfreq() * (1.0 - self._readvelocityaxis() / sc.c)

    def force_relative_offset_axes(self):
        """Force the use of relative offset axes."""
        dx = (self.xaxis.max() - self.xaxis.min()) / 2.0
        self.xaxis = np.linspace(dx, -dx, self.xaxis.size)
        dy = (self.yaxis.max() - self.yaxis.min()) / 2.0
        self.yaxis = np.linspace(-dy, dy, self.yaxis.size)

    # -- UNIT CONVERSIONS -- #

    def jybeam_to_Tb_RJ(self, data=None, nu=None):
        """[Jy/beam] to [K] conversion using Rayleigh-Jeans approximation."""
        nu = self.nu0 if nu is None else nu
        data = self.data if data is None else data
        jy2k = 1e-26 * sc.c**2 / nu**2 / 2. / sc.k
        return jy2k * data / self._calculate_beam_area_str()

    def jybeam_to_Tb(self, data=None, nu=None):
        """[Jy/beam] to [K] conversion using the full Planck law."""
        nu = self.nu0 if nu is None else nu
        data = self.data if data is None else data
        Tb = 1e-26 * abs(data) / self._calculate_beam_area_str()
        Tb = 2.0 * sc.h * nu**3 / Tb / sc.c**2
        Tb = sc.h * nu / sc.k / np.log(Tb + 1.0)
        return np.where(data >= 0.0, Tb, -Tb)

    def Tb_to_jybeam_RJ(self, data=None, nu=None):
        """[K] to [Jy/beam] conversion using Rayleigh-Jeans approxmation."""
        nu = self.nu0 if nu is None else nu
        data = self.data if data is None else data
        jy2k = 1e-26 * sc.c**2 / nu**2 / 2. / sc.k
        return data * self._calculate_beam_area_str() / jy2k

    def Tb_to_jybeam(self, data=None, nu=None):
        """[K] to [Jy/beam] conversion using the full Planck law."""
        nu = self.nu0 if nu is None else nu
        data = self.data if data is None else data
        Fnu = 2. * sc.h * nu**3 / sc.c**2
        Fnu /= np.exp(sc.h * nu / sc.k / abs(data)) - 1.0
        Fnu *= self._calculate_beam_area_str() / 1e-26
        return np.where(data >= 0.0, Fnu, -Fnu)

    # -- BEAM FUNCTIONS -- #

    def _calculate_beam_area_arcsec(self):
        """Beam area in square arcseconds."""
        omega = self.bmin * self.bmaj
        if self.bmin == self.dpix and self.bmaj == self.dpix:
            return omega
        return np.pi * omega / 4. / np.log(2.)

    def _calculate_beam_area_str(self):
        """Beam area in steradians."""
        omega = np.radians(self.bmin / 3600.)
        omega *= np.radians(self.bmaj / 3600.)
        if self.bmin == self.dpix and self.bmaj == self.dpix:
            return omega
        return np.pi * omega / 4. / np.log(2.)

    def _beamkernel(self):
        """Returns the 2D Gaussian kernel for convolution."""
        from astropy.convolution import Gaussian2DKernel
        bmaj = self.bmaj / self.dpix / self.fwhm
        bmin = self.bmin / self.dpix / self.fwhm
        return Gaussian2DKernel(bmin, bmaj, np.radians(self.bpa))

    @staticmethod
    def _convolve_image(image, kernel, fast=True):
        """Convolve the image with the provided kernel."""
        if fast:
            from astropy.convolution import convolve_fft
            return convolve_fft(image, kernel, preserve_nan=True)
        from astropy.convolution import convolve
        return convolve(image, kernel, preserve_nan=True)

    # -- DIAGNOSTIC FUNCTIONS -- #

    def estimate_cube_RMS(self, N=10, r_in=0.0, r_out=1e10):
        """
        Estimate RMS of the cube based on first and last `N` channels and a
        circular area described by an inner and outer radius.

        Args:
            N (int): Number of edge channels to include.
            r_in (float): Inner edge of pixels to consider in [arcsec].
            r_out (float): Outer edge of pixels to consider in [arcsec].

        Returns:
            RMS (float): The RMS based on the requested pixel range.
        """
        r_dep = np.hypot(self.xaxis[None, :], self.yaxis[:, None])
        rmask = np.logical_and(r_dep >= r_in, r_dep <= r_out)
        rms = np.concatenate([self.data[:int(N)], self.data[-int(N):]])
        rms = np.where(rmask[None, :, :], rms, np.nan)
        return np.sqrt(np.nansum(rms**2) / np.sum(np.isfinite(rms)))

    def integrated_spectrum(self, x0=0.0, y0=0.0, inc=0.0, PA=0.0, r_min=None,
                            r_max=None):
        """
        Returns the integrated spectrum over a specified region.

        Args:
            x0 (Optional[float]): Right Ascension offset in [arcsec].
            y0 (Optional[float]): Declination offset in [arcsec].
            inc (Optional[float]): Disk inclination in [deg].
            PA (Optional[float]): Disk position angle in [deg].
            r_min (Optional[float]): Radius to integrate out from in [arcsec].
            r_max (Optional[float]): Radius to integrate out to in [arcsec].

        Returns:
            spectrum, uncertainty (array, array): Something about these.
        """
        rr = self.disk_coords(x0=x0, y0=y0, inc=inc, PA=PA)[0]
        r_max = rr.max() if r_max is None else r_max
        r_min = 0.0 if r_min is None else r_min
        mask = np.logical_and(rr <= r_max, rr >= r_min)
        nbeams = np.where(mask, 1, 0).sum() / self.pix_per_beam
        spectrum = np.array([np.nansum(c[mask]) for c in self.data])
        spectrum *= self.beams_per_pix
        uncertainty = np.sqrt(nbeams) * self.estimate_cube_RMS()
        return spectrum, uncertainty

    # -- PLOTTING FUNCTIONS -- #

    @staticmethod
    def cmap():
        import matplotlib.colors as mcolors
        c2 = plt.cm.Reds(np.linspace(0, 1, 16))
        c1 = plt.cm.Blues_r(np.linspace(0, 1, 16))
        colors = np.vstack((c1, np.ones((2, 4)), c2))
        return mcolors.LinearSegmentedColormap.from_list('eddymap', colors)

    @property
    def extent(self):
        """Cube field of view for use with Matplotlib's ``imshow``."""
        return [self.xaxis[0]+self.dpix/2.0, self.xaxis[-1]-self.dpix/2.0,
                self.yaxis[0]-self.dpix/2.0, self.yaxis[-1]+self.dpix/2.0]

    @property
    def FOV(self):
        """Cube field of view."""
        xFOV = self.xaxis[0] - self.xaxis[-1]
        yFOV = self.yaxis[-1] - self.yaxis[0]
        return np.mean([xFOV, yFOV])

    def extent_au(self, dist=1.0):
        """Cube field of view in [au] for use with Matplotlib's ``imshow``."""
        return dist * np.squeeze(self.extent)

    def plot_beam(self, ax, x0=0.1, y0=0.1, **kwargs):
        """Plot the sythensized beam on the provided axes."""
        from matplotlib.patches import Ellipse
        beam = Ellipse(ax.transLimits.inverted().transform((x0, y0)),
                       width=self.bmin, height=self.bmaj, angle=-self.bpa,
                       fill=kwargs.get('fill', False),
                       hatch=kwargs.get('hatch', '//////////'),
                       lw=kwargs.get('linewidth', kwargs.get('lw', 1)),
                       color=kwargs.get('color', kwargs.get('c', 'k')),
                       zorder=kwargs.get('zorder', 1000))
        ax.add_patch(beam)

    def _gentrify_plot(self, ax):
        """Gentrify the plot with a grid, label axes and a beam."""
        ax.set_aspect(1)
        ax.grid(ls='--', color='k', alpha=0.2, lw=0.5)
        ax.tick_params(which='both', right=True, top=True)
        ax.set_xlim(self.xaxis.max(), self.xaxis.min())
        ax.set_ylim(self.yaxis.min(), self.yaxis.max())
        ax.xaxis.set_major_locator(MaxNLocator(5, min_n_ticks=3))
        ax.yaxis.set_major_locator(MaxNLocator(5, min_n_ticks=3))
        ticks = np.diff(ax.xaxis.get_majorticklocs()).mean() / 5.0
        ax.xaxis.set_minor_locator(MultipleLocator(ticks))
        ax.yaxis.set_minor_locator(MultipleLocator(ticks))
        ax.set_xlabel('Offset (arcsec)')
        ax.set_ylabel('Offset (arcsec)')
        if self.bmaj is not None:
            self.plot_beam(ax=ax)

    def plot_maximum(self, ax=None, imshow_kwargs=None, return_fig=False):
        """
        Plot the maximum value along each spectrum.

        Args:
            ax (Optional[matplotlib axis]): Axis used for the plotting.
            imshow_kwargs (Optional[dict]): Kwargs to pass to
                ``matplotlib.imshow``.
            return_fig (Optional[bool]): Whether to return the figure instance.
                If an axis was provided, this will always be ``False``.

        Returns:
            fig (matplotlib figure): If ``return_fig=True``, will return the
                figure for continued plotting.
        """

        # Dummy axis to overplot.

        if ax is None:
            fig, ax = plt.subplots()
        else:
            return_fig = False

        # Collapse the data.

        dmax = np.nanmax(self.data, axis=0)
        vmax = np.percentile(dmax, [98])

        # Imshow defaults and plot the figure.

        imshow_kwargs = {} if imshow_kwargs is None else imshow_kwargs
        imshow_kwargs['interpolation'] = 'nearest'
        imshow_kwargs['extent'] = self.extent
        imshow_kwargs['origin'] = 'lower'
        imshow_kwargs['cmap'] = imshow_kwargs.pop('cmap', 'turbo')
        imshow_kwargs['vmin'] = imshow_kwargs.pop('vmin', 0.0)
        imshow_kwargs['vmax'] = imshow_kwargs.pop('vmax', vmax)

        im = ax.imshow(dmax, **imshow_kwargs)
        cb = plt.colorbar(im, ax=ax, pad=0.03, extend='both')
        cb.set_label('Peak Intensity (Jy/beam)', rotation=270, labelpad=13)
        cb.minorticks_on()
        self._gentrify_plot(ax=ax)

        # Returns

        if return_fig:
            return fig

    def plot_surface(self, x0=0.0, y0=0.0, inc=0.0, PA=0.0, z0=None, psi=None,
                     r_cavity=None, r_taper=None, q_taper=None, w_i=None,
                     w_r=None, w_t=None, z_func=None, w_func=None,
                     shadowed=False, r_max=None, mask=None, fill=None, ax=None,
                     contour_kwargs=None, imshow_kwargs=None, return_fig=True,
                     **_):
        """
        Overplot the emission surface onto the provided axis.

        Args:
            ax (Optional[AxesSubplot]): Axis to plot onto.
            x0 (Optional[float]): Source right ascension offset [arcsec].
            y0 (Optional[float]): Source declination offset [arcsec].
            inc (Optional[float]): Source inclination [deg].
            PA (Optional[float]): Source position angle [deg]. Measured
                between north and the red-shifted semi-major axis in an
                easterly direction.
            z0 (Optional[float]): Aspect ratio at 1" for the emission surface.
                To get the far side of the disk, make this number negative.
            psi (Optional[float]): Flaring angle for the emission surface.
            z1 (Optional[float]): Aspect ratio correction term at 1" for the
                emission surface. Should be opposite sign to z0.
            phi (Optional[float]): Flaring angle correction term for the
                emission surface.
            w_i (Optional[float]): Warp inclination in [degrees] at the disk
                center.
            w_r (Optional[float]): Scale radius of the warp in [arcsec].
            w_t (Optional[float]): Angle of nodes of the warp in [degrees].
            r_max (Optional[float]): Outer radius to plot.
            mask (Optional[array]): A 2D mask to define where the surcace is
                plotted.
            ntheta (Optional[int]): Number of theta contours to plot.
            nrad (Optional[int]): Number of radial contours to plot.
            mask (Optional[array]): Mask used to define regions where the
                surface should be plot. If not provided, will plot everywhere.
            check_mask (Optional[bool]): Mask regions which are like projection
                errors for highly flared surfaces.

        Returns:
            ax (AxesSubplot): Axis with the contours overplotted.
        """

        # Dummy axis to overplot.

        if ax is None:
            fig, ax = plt.subplots()
        else:
            return_fig = False

        rvals, tvals, zvals = self.disk_coords(x0=x0, y0=y0,
                                               inc=inc, PA=PA,
                                               z0=z0, psi=psi,
                                               r_cavity=r_cavity,
                                               r_taper=r_taper,
                                               q_taper=q_taper,
                                               w_i=w_i, w_r=w_r, w_t=w_t,
                                               z_func=z_func,
                                               w_func=w_func,
                                               shadowed=shadowed)

        # Mask the data based on r_max.

        r_max = np.nanmax(rvals) if r_max is None else r_max
        zvals = np.where(rvals <= r_max, zvals, np.nan)
        tvals = np.where(rvals <= r_max, tvals, np.nan)
        tvals = np.where(rvals >= 0.5 * self.bmaj, tvals, np.nan)
        rvals = np.where(rvals <= r_max, rvals, np.nan)

        # Mask the data based on user-defined mask.

        if mask is not None:
            rvals = np.where(mask, rvals, np.nan)
            tvals = np.where(mask, tvals, np.nan)
            zvals = np.where(mask, zvals, np.nan)

        # Fill in the background.

        if fill is not None:
            kw = {} if imshow_kwargs is None else imshow_kwargs
            kw['origin'] = 'lower'
            kw['extent'] = self.extent
            ax.imshow(eval(fill), **kw)

        # Draw the contours. The azimuthal angles are drawn on individually to
        # avoid having overlapping lines about the +\- pi boundary making a
        # particularly thick line.

        kw = {} if contour_kwargs is None else contour_kwargs
        kw['levels'] = kw.pop('levels', np.arange(0.5, 0.99 * r_max, 0.5))
        kw['levels'] = np.append(kw['levels'], 0.99 * r_max)
        kw['linewidths'] = kw.pop('linewidths', 1.0)
        kw['colors'] = kw.pop('colors', 'k')
        ax.contour(self.xaxis, self.yaxis, rvals, **kw)

        kw['levels'] = [0.0]
        for t in np.arange(-np.pi, np.pi, np.pi / 8.0):
            if t - 0.1 < -np.pi:
                a = np.where(abs(tvals - t) <= 0.4,
                             tvals - t, np.nan)
                b = np.where(abs(tvals - 2.0 * np.pi - t) <= 0.4,
                             tvals - 2.0 * np.pi - t, np.nan)
                amask = np.where(np.isfinite(a), 1, -1)
                bmask = np.where(np.isfinite(b), 1, -1)
                a = np.where(np.isfinite(a), a, 0.0)
                b = np.where(np.isfinite(b), b, 0.0)
                ttmp = np.where(amask * bmask < 1, a + b, np.nan)
            elif t + 0.1 > np.pi:
                a = np.where(abs(tvals - t) <= 0.4,
                             tvals - t, np.nan)
                b = np.where(abs(tvals + 2.0 * np.pi - t) <= 0.4,
                             tvals + 2.0 * np.pi - t, np.nan)
                amask = np.where(np.isfinite(a), 1, -1)
                bmask = np.where(np.isfinite(b), 1, -1)
                a = np.where(np.isfinite(a), a, 0.0)
                b = np.where(np.isfinite(b), b, 0.0)
                ttmp = np.where(amask * bmask < 1, a + b, np.nan)
            else:
                ttmp = np.where(abs(tvals - t) <= 0.4, tvals - t, np.nan)
            ax.contour(self.xaxis, self.yaxis, ttmp, **kw)
        ax.set_xlim(max(ax.get_xlim()), min(ax.get_xlim()))
        ax.set_aspect(1)

        if return_fig:
            return fig

    def plot_spectrum(self, ax=None, x0=0.0, y0=0.0, inc=0.0, PA=0.0,
                      r_min=None, r_max=None, return_fig=False):
        """
        Plot the integrated spectrum.

        Args:
            x0 (Optional[float]): Right Ascension offset in [arcsec].
            y0 (Optional[float]): Declination offset in [arcsec].
            inc (Optional[float]): Disk inclination in [deg].
            PA (Optional[float]): Disk position angle in [deg].
            r_max (Optional[float]): Radius to integrate out to in [arcsec].
        """
        if ax is None:
            fig, ax = plt.subplots()
        else:
            return_fig = False
        x = self.velax.copy() / 1e3
        y, dy = self.integrated_spectrum(x0, y0, inc, PA, r_min, r_max)
        ax.axhline(0.0, ls='--', lw=1.0, color='0.9', zorder=-9)
        ax.step(x, y, where='mid', lw=1.0, color='k')
        ax.errorbar(x, y, dy, fmt=' ', lw=1.0, color='k', zorder=-8)
        ax.set_xlabel("Velocity (km/s)")
        ax.set_ylabel("Integrated Flux (Jy)")
        ax.set_xlim(x[0], x[-1])
        ticks = np.diff(ax.xaxis.get_majorticklocs()).mean() / 5.0
        ax.xaxis.set_minor_locator(MultipleLocator(ticks))
        ticks = np.diff(ax.yaxis.get_majorticklocs()).mean() / 5.0
        ax.yaxis.set_minor_locator(MultipleLocator(ticks))

        if return_fig:
            return fig



# -*- coding: utf-8 -*-

from .datacube import datacube
from .annulus import annulus
import numpy as np
import warnings

warnings.filterwarnings("ignore")


class linecube(datacube):
    """
    Read in a line cube and initialize the class.

    Args:
        path (str): Relative path to the rotation map you want to fit.
        FOV (Optional[float]): If specified, clip the data down to a
            square field of view with sides of `FOV` [arcsec].
        fill (Optional[float/None]): Value to fill any NaN values.
        velocity_range (Optional[list]): A velocity range in [m/s] to
            clip the data down to: ``[min_velo, max_velo]``.
    """

    def __init__(self, path, FOV=None, fill=0.0, velocity_range=None):
        datacube.__init__(self, path=path, FOV=FOV, fill=fill,
                          velocity_range=velocity_range)

    # -- ROTATION PROFILE FUNCTIONS -- #

    def get_velocity_profile(self, rbins=None, fit_method='GP', fit_vrad=False,
                             x0=0.0, y0=0.0, inc=0.0, PA=0.0, z0=0.0,
                             psi=1.0, r_cavity=0.0, r_taper=np.inf,
                             q_taper=1.0, w_i=None, w_r=None, w_t=None,
                             z_func=None, shadowed=False, phi_min=None,
                             phi_max=None, exclude_phi=False, abs_phi=False,
                             mask_frame='disk', user_mask=None,
                             beam_spacing=True, deproject=False, niter=1,
                             get_vlos_kwargs=None, weighted_average=True,
                             return_samples=False):
        """
        Returns the rotational and, optionally, radial velocity profiles under
        the assumption that the disk is azimuthally symmetric (at least across
        the regions extracted).

        Several different inference methods can be used through the
        ``fit_method`` argument:

            - ``'GP'``:  Models the aligned and stacked spectrum as a Gaussian
                         Process to remain agnostic about the underlying true
                         line profile. This is the default.
            - ``'dV'``:  Minimize the line width of the aligned and stacked
                         spectrum. Assumes the underyling profile is a
                         Gaussian.
            - ``'SNR'``: Maximizes the SNR of the aligned and stacked spectrum.
                         Assumes the underlying profile is a Gaussian.
            - ``'SHO'``: Fits the azimuthal dependence of the line centroids
                         with a simple harmonic oscillator. Requires a choice
                         of ``centroid_method`` to determine the line centers.

        By default, these methods are applied to annuli across the whole radial
        and azimuthal range of the disk. Masks can be adopted, either limiting
        or excluding azimuthal regions, or a user-defined mask can be provided.

        Boostrapping is possible (multiple iterations using random draws) to
        estimate the uncertainties on the velocities which may be
        underestimated with a single interation.

        Args:
            rbins (Optional[arr]): Array of bin edges of the annuli.
            fit_method (Optional[str]): Method used to infer the velocities.
            fit_vrad (Optional[bool]): Whether to include radial velocities in
                the fit.
            x0 (Optional[float]): Source right ascension offset [arcsec].
            y0 (Optional[float]): Source declination offset [arcsec].
            inc (Optional[float]): Source inclination [degrees]. A positive
                inclination denotes a disk rotating clockwise on the sky, while
                a negative inclination represents a counter-clockwise rotation.
            PA (Optional[float]): Source position angle [degrees]. Measured
                between north and the red-shifted semi-major axis in an
                easterly direction.
            z0 (Optional[float]): Aspect ratio at 1" for the emission surface.
                To get the far side of the disk, make this number negative.
            psi (Optional[float]): Flaring angle for the emission surface.
            z1 (Optional[float]): Correction term for ``z0``.
            phi (Optional[float]): Flaring angle correction term for the
                emission surface.
            r_cavity (Optional[float]): Outer radius of a cavity. Within this
                region the emission surface is taken to be zero.
            w_i (Optional[float]): Warp inclination in [degrees] at the disk
                center.
            w_r (Optional[float]): Scale radius of the warp in [arcsec].
            w_t (Optional[float]): Angle of nodes of the warp in [degrees].
            z_func=None,
            shadowed=False,
            phi_min=None,
            phi_max=None,
            exclude_phi=False,
            abs_phi=False,
            mask_frame='disk',
            user_mask=None,
            beam_spacing=True,
            deproject=False,
            niter=1,
            get_vlos_kwargs=None,
            weighted_average=True,
            return_samples=False

        Returns:
            TBD
        """

        if niter == 0:
            raise ValueError("`niter` must be >= 1.")

        # Single iteration.

        if niter == 1:
            return self._velocity_profile(rbins=rbins,
                                          fit_method=fit_method,
                                          fit_vrad=fit_vrad,
                                          x0=x0,
                                          y0=y0,
                                          inc=inc,
                                          PA=PA,
                                          z0=z0,
                                          psi=psi,
                                          r_cavity=r_cavity,
                                          r_taper=r_taper,
                                          q_taper=q_taper,
                                          w_i=w_i,
                                          w_r=w_r,
                                          w_t=w_t,
                                          z_func=z_func,
                                          shadowed=shadowed,
                                          phi_min=phi_min,
                                          phi_max=phi_max,
                                          exclude_phi=exclude_phi,
                                          abs_phi=abs_phi,
                                          mask_frame=mask_frame,
                                          user_mask=user_mask,
                                          beam_spacing=beam_spacing,
                                          get_vlos_kwargs=get_vlos_kwargs)

        # Multiple iterations.

        if beam_spacing is False:
            raise ValueError("niter must equal 1 when beam_spacing=False.")

        samples = [self._velocity_profile(rbins=rbins,
                                          fit_method=fit_method,
                                          fit_vrad=fit_vrad,
                                          x0=x0,
                                          y0=y0,
                                          inc=inc,
                                          PA=PA,
                                          z0=z0,
                                          psi=psi,
                                          r_cavity=r_cavity,
                                          r_taper=r_taper,
                                          q_taper=q_taper,
                                          w_i=w_i,
                                          w_r=w_r,
                                          w_t=w_t,
                                          z_func=z_func,
                                          shadowed=shadowed,
                                          phi_min=phi_min,
                                          phi_max=phi_max,
                                          exclude_phi=exclude_phi,
                                          abs_phi=abs_phi,
                                          mask_frame=mask_frame,
                                          user_mask=user_mask,
                                          beam_spacing=beam_spacing,
                                          get_vlos_kwargs=get_vlos_kwargs)
                   for _ in range(niter)]

        # Just return the samples.

        if return_samples:
            return samples
        rpnts = samples[0][0]
        profiles = np.array([s[1] for s in samples])

        # Calculate weights.

        if weighted_average:
            weights = [1.0 / s[2] for s in samples]
            weights = np.where(np.isfinite(weights), weights, 0.0)
        else:
            weights = np.ones(profiles.shape)
        if np.all(np.sum(weights, axis=0) == 0.0):
            weights = np.ones(profiles.shape)
        M = np.sum(weights != 0.0, axis=0)

        # Weighted average.

        profile = np.average(profiles, weights=weights, axis=0)

        # Weighted standard deviation.

        uncertainty = weights * (profiles - profile[None, :, :])**2
        uncertainty = np.sum(uncertainty, axis=0)
        uncertainty /= (M - 1.0) / M * np.sum(weights, axis=0)
        uncertainty = np.sqrt(uncertainty)
        return rpnts, profile, uncertainty

    def _velocity_profile(self, rbins=None, fit_method='GP', fit_vrad=False,
                          x0=0.0, y0=0.0, inc=0.0, PA=0.0, z0=0.0,
                          psi=1.0, r_cavity=0.0, r_taper=np.inf, q_taper=1.0,
                          w_i=None, w_r=None, w_t=None, z_func=None,
                          shadowed=False, phi_min=None, phi_max=None,
                          exclude_phi=False, abs_phi=False, mask_frame='disk',
                          user_mask=None, beam_spacing=True, deproject=False,
                          get_vlos_kwargs=None):
        """Returns the velocity (rotational and radial) profiles."""

        # Define the radial binning.

        if rbins is None:
            rbins = np.arange(0, self.xaxis.max(), 0.25 * self.bmaj)
        rpnts = np.mean([rbins[1:], rbins[:-1]], axis=0)

        # Set up the kwargs for the fitting.

        kw = {} if get_vlos_kwargs is None else get_vlos_kwargs
        kw['fit_vrad'] = kw.pop('fit_vrad', fit_vrad)
        kw['fit_method'] = fit_method

        # Cycle through the annuli.

        profiles = []
        uncertainties = []
        for r_min, r_max in zip(rbins[:-1], rbins[1:]):
            annulus = self.get_annulus(r_min=r_min,
                                       r_max=r_max,
                                       phi_min=phi_min,
                                       phi_max=phi_max,
                                       exclude_phi=exclude_phi,
                                       abs_phi=abs_phi,
                                       x0=x0,
                                       y0=y0,
                                       inc=inc,
                                       PA=PA,
                                       z0=z0,
                                       psi=psi,
                                       r_cavity=r_cavity,
                                       r_taper=r_taper,
                                       q_taper=q_taper,
                                       w_i=w_i,
                                       w_r=w_r,
                                       w_t=w_t,
                                       z_func=z_func,
                                       shadowed=shadowed,
                                       mask_frame=mask_frame,
                                       user_mask=user_mask,
                                       beam_spacing=beam_spacing)
            output = annulus.get_vlos(**kw)

            # Parse the outputs.

            par = 2 if fit_vrad else 1
            output = np.empty(par) if np.all(np.isnan(output)) else output
            if fit_method.lower() in ['dv', 'snr']:
                profiles += [output]
                uncertainties += [np.ones(par) * np.nan]
            elif fit_method.lower() == 'sho':
                profiles += [output[0]]
                uncertainties += [output[1]]
            elif fit_method.lower() == 'gp':
                profiles += [output[0]]
                uncertainties += [output[1]]

        # Make sure the returned arrays are in the (nparam, nrad) form.

        profiles = np.atleast_2d(profiles)
        uncertainties = np.atleast_2d(uncertainties)
        if profiles.shape[0] == rpnts.size:
            profiles = profiles.T
            uncertainties = uncertainties.T

        # Deproject the velocity profiles to account for the inclination.

        if deproject:
            profiles /= np.sin(np.radians(abs(inc)))
            uncertainties /= np.sin(np.radians(abs(inc)))

        return rpnts, profiles, uncertainties

    # -- ANNULUS FUNCTIONS -- #

    def get_annulus(self, r_min, r_max, phi_min=None, phi_max=None,
                    exclude_phi=False, abs_phi=False, x0=0.0, y0=0.0, inc=0.0,
                    PA=0.0, z0=0.0, psi=1.0, r_cavity=0.0, r_taper=np.inf,
                    q_taper=1.0, w_i=None, w_r=None, w_t=None, z_func=None,
                    shadowed=False, mask_frame='disk', user_mask=None,
                    beam_spacing=True, annulus_kwargs=None):
        """
        Returns an annulus instance.
        """

        # Calculate and flatten the mask.

        mask = self.get_mask(r_min=r_min,
                             r_max=r_max,
                             phi_min=phi_min,
                             phi_max=phi_max,
                             exclude_phi=exclude_phi,
                             abs_phi=abs_phi,
                             x0=x0,
                             y0=y0,
                             inc=inc,
                             PA=PA,
                             z0=z0,
                             psi=psi,
                             r_cavity=r_cavity,
                             r_taper=r_taper,
                             q_taper=q_taper,
                             w_i=w_i,
                             w_r=w_r,
                             w_t=w_t,
                             z_func=z_func,
                             shadowed=shadowed,
                             mask_frame=mask_frame,
                             user_mask=user_mask)
        if mask.shape != self.data[0].shape:
            raise ValueError("mask is incorrect shape: {}.".format(mask.shape))
        mask = mask.flatten()

        # Flatten the data and get the deprojected pixel coordinates.

        dvals = self.data.copy().reshape(self.data.shape[0], -1)
        rvals, pvals = self.disk_coords(x0=x0,
                                        y0=y0,
                                        inc=inc,
                                        PA=PA,
                                        z0=z0,
                                        psi=psi,
                                        r_cavity=r_cavity,
                                        r_taper=r_taper,
                                        q_taper=q_taper,
                                        w_i=w_i,
                                        w_r=w_r,
                                        w_t=w_t,
                                        z_func=z_func,
                                        shadowed=shadowed)[:2]
        rvals, pvals = rvals.flatten(), pvals.flatten()
        dvals, rvals, pvals = dvals[:, mask].T, rvals[mask], pvals[mask]

        # Thin down to spatially independent pixels.

        rvals, pvals, dvals = self._independent_samples(beam_spacing,
                                                        rvals, pvals, dvals)

        # Return the annulus instance.

        annulus_kwargs = {} if annulus_kwargs is None else annulus_kwargs
        return annulus(spectra=dvals, pvals=pvals, velax=self.velax, inc=inc,
                       **annulus_kwargs)

    def _independent_samples(self, beam_spacing, rvals, pvals, dvals):
        """Returns spatially independent samples."""

        if not beam_spacing:
            return rvals, pvals, dvals

        # Order pixels in increasing phi.

        idxs = np.argsort(pvals)
        dvals, pvals = dvals[idxs], pvals[idxs]

        # Calculate the sampling rate.

        sampling = float(beam_spacing) * self.bmaj
        sampling /= np.mean(rvals) * np.median(np.diff(pvals))
        sampling = np.floor(sampling).astype('int')

        # If the sampling rate is above 1, start at a random location in
        # the array and sample at this rate, otherwise don't sample. This
        # happens at small radii, for example.

        if sampling > 1:
            start = np.random.randint(0, pvals.size)
            rvals = np.concatenate([rvals[start:], rvals[:start]])
            pvals = np.concatenate([pvals[start:], pvals[:start]])
            dvals = np.vstack([dvals[start:], dvals[:start]])
            rvals = rvals[::sampling]
            pvals = pvals[::sampling]
            dvals = dvals[::sampling]
        else:
            print("Pixels appear to be close to spatially independent.")

        return rvals, pvals, dvals

    # -- PLOTTING FUNCTIONS -- #

    def plot_mask(self, ax, r_min=None, r_max=None, exclude_r=False,
                  phi_min=None, phi_max=None, exclude_phi=False, abs_phi=False,
                  mask_frame='disk', mask=None, x0=0.0, y0=0.0, inc=0.0,
                  PA=0.0, z0=0.0, psi=1.0, r_cavity=None, r_taper=None,
                  q_taper=None, w_i=None, w_r=None, w_t=None, z_func=None,
                  mask_color='k', mask_alpha=0.5, contour_kwargs=None,
                  contourf_kwargs=None, shadowed=False):
        """
        Plot the boolean mask on the provided axis to check that it makes
        sense.

        Args:
            ax (matplotib axis instance): Axis to plot the mask.
            r_min (Optional[float]): Minimum midplane radius of the annulus in
                [arcsec]. Defaults to minimum deprojected radius.
            r_max (Optional[float]): Maximum midplane radius of the annulus in
                [arcsec]. Defaults to the maximum deprojected radius.
            exclude_r (Optional[bool]): If ``True``, exclude the provided
                radial range rather than include.
            PA_min (Optional[float]): Minimum polar angle of the segment of the
                annulus in [degrees]. Note this is the polar angle, not the
                position angle.
            PA_max (Optional[float]): Maximum polar angle of the segment of the
                annulus in [degrees]. Note this is the polar angle, not the
                position angle.
            exclude_PA (Optional[bool]): If ``True``, exclude the provided
                polar angle range rather than include it.
            abs_PA (Optional[bool]): If ``True``, take the absolute value of
                the polar angle such that it runs from 0 [deg] to 180 [deg].
            x0 (Optional[float]): Source center offset along the x-axis in
                [arcsec].
            y0 (Optional[float]): Source center offset along the y-axis in
                [arcsec].
            inc (Optional[float]): Inclination of the disk in [degrees].
            PA (Optional[float]): Position angle of the disk in [degrees],
                measured east-of-north towards the redshifted major axis.
            z0 (Optional[float]): Emission height in [arcsec] at a radius of
                1".
            psi (Optional[float]): Flaring angle of the emission surface.
            z1 (Optional[float]): Correction to emission height at 1" in
                [arcsec].
            phi (Optional[float]): Flaring angle correction term.
            z_func (Optional[function]): A function which provides
                :math:`z(r)`. Note that no checking will occur to make sure
                this is a valid function.
            mask_color (Optional[str]): Color used for the mask lines.
            mask_alpha (Optional[float]): The alpha value of the filled contour
                of the masked regions. Setting ``mask_alpha=0.0`` will remove
                the filling.

            contour_kwargs (Optional[dict]): Kwargs to pass to contour for
                drawing the mask.

        Returns:
            ax : The matplotlib axis instance.
        """
        # Grab the mask.
        if mask is None:
            mask = self.get_mask(r_min=r_min, r_max=r_max, exclude_r=exclude_r,
                                 phi_min=phi_min, phi_max=phi_max,
                                 exclude_phi=exclude_phi, abs_phi=abs_phi,
                                 mask_frame=mask_frame, x0=x0, y0=y0, inc=inc,
                                 PA=PA, z0=z0, psi=psi, r_cavity=r_cavity,
                                 r_taper=r_taper, q_taper=q_taper, w_i=w_i,
                                 w_r=w_r, w_t=w_t, z_func=z_func,
                                 shadowed=shadowed)
        assert mask.shape[0] == self.yaxis.size, "Wrong y-axis shape for mask."
        assert mask.shape[1] == self.xaxis.size, "Wrong x-axis shape for mask."

        # Set the default plotting style.

        contour_kwargs = {} if contour_kwargs is None else contour_kwargs
        contour_kwargs['colors'] = contour_kwargs.pop('colors', mask_color)
        contour_kwargs['linewidths'] = contour_kwargs.pop('linewidths', 1.0)
        contour_kwargs['linestyles'] = contour_kwargs.pop('linestyles', '-')
        contourf_kwargs = {} if contourf_kwargs is None else contourf_kwargs
        contourf_kwargs['alpha'] = contourf_kwargs.pop('alpha', mask_alpha)
        contourf_kwargs['colors'] = contourf_kwargs.pop('colors', mask_color)

        # Plot the contour and return the figure.

        ax.contourf(self.xaxis, self.yaxis, mask, [-.5, .5], **contourf_kwargs)
        ax.contour(self.xaxis, self.yaxis, mask, 1, **contour_kwargs)

    def get_spectrum(self, coords, x0=0.0, y0=0.0, inc=0.0, PA=0.0,
                     z0=0.0, psi=1.0, z_func=None, frame='sky',
                     coord_type='cartesian', area=0.0, beam_weighting=False,
                     return_mask=False):
        """
        Return a spectrum at a position defined by a coordinates given either
        in sky-frame position (``frame='sky'``) or a disk-frame location
        (``frame='disk'``). The coordinates can be either in a cartesian system
        (``coord_type='cartesian'``) or cylindrical system
        (``coord_type='cylindrical'``).

        By default the returned spectrum is extracted at the pixel closest to
        the provided coordinates. If ``area`` is set to a positive value, then
        a beam-shaped area is averaged over, where ``area`` sets the size of
        this region in number of beams. For example ``area=2.0`` will result
        in an average over an area twice the size of the beam.

        If an area is averaged over, you can also weight the pixels by the beam
        response with ``beam_weighting=True``. This will reduce the weight of
        pixels that are further away from the beam center.

         Finally, to check that you're extracting what you think you are, you
         can return the mask (and weights) used for the extraction with
         ``return_mask=True``. Note that if ``beam_weighting=False`` then all
         ``weights`` will be 1.

         TODO: Check that the returned uncertainties are reasonable.

        Args:
            coords (tuple): The coordinates from where you want to extract a
                spectrum. Must be a length 2 tuple.
            x0 (Optional[float]): RA offset in [arcsec].
            y0 (Optional[float]): Dec offset in [arcsec].
            inc (Optional[float]): Inclination of source in [deg]. Only
                required for ``frame='disk'``.
            PA (Optional[float]): Position angle of source in [deg]. Only
                required for ``frame='disk'``.
            frame (Optional[str]): The frame that the ``coords`` are given.
                Either ``'disk'`` or ``'sky'``.
            coord_type (Optional[str]): The type of coordinates given, either
                ``'cartesian'`` or ``'cylindrical'``.
            area (Optional[float]): The area to average over in units of the
                beam area. Note that this take into account the beam aspect
                ratio and position angle. For a single pixel extraction use
                ``area=0.0``.
            beam_weighting (Optional[bool]): Whether to use the beam response
                function to weight the averaging of the spectrum.
            return_mask (Optional[bool]): Whether to return the mask and
                weights used to extract the spectrum.

        Retuns (if ``return_mask=False``):
            x, y, dy (arrays): The velocity axis, extracted spectrum and
            associated uncertainties.
        (if ``return_mask=True``):
            mask, weights (arrays): Arrays of the mask used to extract the
            spectrum and the weighted used for the averaging.
        """

        # TODO:
        #   1 - Check if it's three coordinates, or just two.
        #   2 - Check which frame it's in.

        # Convert the input coordinate into on-sky cartesian coordinates
        # relative to the center of the image.

        if frame.lower() == 'sky':
            if inc != 0.0 or PA != 0.0:
                message = "WARNING: You shouldn't need to specify `inc` or "
                message += "`PA` when using `frame='sky'`."
                print(message)
            c1 = np.squeeze(coords[0])
            c2 = np.squeeze(coords[1])
            if coord_type.lower() == 'cartesian':
                x, y = c1 + x0, c2 + y0
            elif coord_type.lower() == 'cylindrical':
                x = x0 + c1 * np.cos(c2 - np.radians(90.0))
                y = y0 - c1 * np.sin(c2 - np.radians(90.0))
        elif frame.lower() == 'disk':
            x, y = self.disk_to_sky(coords=coords, coord_type=coord_type,
                                    inc=inc, PA=PA, x0=x0, y0=y0)
        assert x.size == y.size == 1

        # Define the area to average over.

        if area == 0.0:
            x_pix = abs(self.xaxis - x).argmin()
            y_pix = abs(self.yaxis - y).argmin()
            mask = np.zeros(self.data[0].shape)
            weights = np.zeros(mask.shape)
            mask[y_pix, x_pix] = 1
            weights[y_pix, x_pix] = 1
        elif area > 0.0:
            mask = self._beam_mask(x, y, stretch=area)
            weights = self._beam_mask(x, y, stretch=area, response=True)
        else:
            raise ValueError("`area` must be a non-negative value.")
        weights = weights if beam_weighting else mask

        # If requested, return the mask and the weighting instead.

        if return_mask:
            return mask, weights

        # Otherwise, extract the spectrum and average it.

        y = [np.average(c * mask, weights=weights) for c in self.data]
        dy = max(1.0, mask.sum() * self.beams_per_pix)**-0.5 * self.rms
        return self.velax, np.array(y), np.array([dy for _ in y])



#! /usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="astro-eddy",
    version="2.2.2",
    author="Richard Teague",
    author_email="rteague@mit.edu",
    description=("Tools to recover expectionally precise rotation curves from "
                 "spatially resolved spectra."),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/richteague/eddy",
    packages=["eddy"],
    license="MIT",
    install_requires=[
        "scipy>=1",
        "numpy",
        "matplotlib>=3",
        "emcee>=3",
        "corner>=2",
        "zeus-mcmc",
        ],
    package_data={'eddy': ['*.yml']},
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3.5",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt
import numpy as np


# -- MCMC / OPTIMIZATION FUNCTIONS -- #

def random_p0(p0, scatter, nwalkers):
    """
    Introduce scatter to starting positions while allowing for starting
    positions scattered about zero.

    Args:
        p0 (list): Starting positions.
        scatter (float): Scatter to apply to the starting positions.
        nwalkers (int): Number of walkers.

    Returns:
        p0 (ndarray): A (nwalkers, ndim) shaped array of starting positions.
    """
    p0 = np.atleast_1d(np.squeeze(p0))
    dp0 = np.random.randn(nwalkers * len(p0)).reshape(nwalkers, len(p0))
    dp0 = np.where(p0 == 0.0, 1.0, p0)[None, :] * (1.0 + scatter * dp0)
    return np.where(p0[None, :] == 0.0, dp0 - 1.0, dp0)


def _errors(x, dy, return_uncertainty):
    """
    Parse the inputs related to errors for use with scipy.optimize.curve_fit.

    Args:
        x (array): Dependent variable.
        dy (array): Uncertainty.
        return_uncertainty (bool): Whether to return the uncertainty.

    Returns:
        dy (array), return_uncertainty (bool), absolute_sigma (bool): Values
            needed for scipy.optimize.curve_fit.

    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    dy = np.ones(x.size) * (1.0 if dy is None else dy)
    if np.all(np.isnan(dy)):
        dy = np.ones(x.size)
        absolute_sigma = False
    else:
        absolute_sigma = True
    return dy, return_uncertainty, absolute_sigma


def fit_gaussian(x, y, dy=None, return_uncertainty=None):
    """
    Fit a Gaussian form to (x, y[, dy]).

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.

    Returns:
        popt (array)[, cvar(array)]: The best fit parameters and, optionally,
        the uncertainty on those parameters.
    """
    dy, return_uncertainty, absolute_sigma = _errors(x, dy, return_uncertainty)
    p0 = get_p0_gaussian(x, y)
    try:
        popt, cvar = curve_fit(gaussian, x, y, sigma=dy, p0=p0,
                               absolute_sigma=absolute_sigma,
                               maxfev=100000)
        cvar = np.diag(cvar)**0.5
    except Exception:
        popt = [np.nan, np.nan, np.nan]
        cvar = popt.copy()
    return (popt, cvar) if return_uncertainty else popt


def fit_gaussian_thick(x, y, dy=None, return_uncertainty=None):
    """
    Fit an optically thick Gaussian function to (x, y[, dy]).

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.

    Returns:
        popt (array)[, cvar(array)]: The best fit parameters and, optionally,
        the uncertainty on those parameters.
    """
    dy, return_uncertainty, absolute_sigma = _errors(x, dy, return_uncertainty)
    p0 = fit_gaussian(x=x, y=y, dy=dy,
                      return_uncertainty=False)
    p0 = np.append(p0, 0.5)
    try:
        popt, cvar = curve_fit(gaussian_thick, x, y, sigma=dy, p0=p0,
                               absolute_sigma=absolute_sigma,
                               maxfev=100000)
        cvar = np.diag(cvar)**0.5
    except Exception:
        popt = [np.nan, np.nan, np.nan, np.nan]
        cvar = popt.copy()
    return (popt, cvar) if return_uncertainty else popt


def fit_double_gaussian(x, y, dy=None, return_uncertainty=None):
    """
    Fit two Gaussian lines to (x, y[, dy]) where the maximum of the two
    Gaussians is used as the model.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.

    Returns:
        popt (array)[, cvar(array)]: The best fit parameters and, optionally,
        the uncertainty on those parameters.
    """

    # Defaults.

    dy, return_uncertainty, absolute_sigma = _errors(x, dy, return_uncertainty)
    popt = [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    cvar = popt.copy()

    # Initial single Gaussian fit.

    popt_a, cvar_a = fit_gaussian(x=x, y=y, dy=dy, return_uncertainty=True)
    if not all(np.isfinite(popt_a)):
        return (popt, cvar) if return_uncertainty else popt

    # Double Gaussian fit. Try first a fit where the profile is a sum of two
    # components, then use this as strong priors for a max of two component
    # fit. This is because the max version is hard to optimize.

    p0 = [popt_a[0] + popt_a[1], popt_a[1], 0.8 * popt_a[2],
          popt_a[0] - popt_a[1], popt_a[1], 0.8 * popt_a[2]]
    try:
        popt_b, _ = curve_fit(double_gaussian_sum, x, y, sigma=dy, p0=p0,
                                   absolute_sigma=absolute_sigma,
                                   maxfev=100000)
        p0 = [popt_b[0], popt_b[1], double_gaussian_sum(popt_b[0], *popt_b),
              popt_b[3], popt_b[4], double_gaussian_sum(popt_b[3], *popt_b)]
        popt_b, cvar_b = curve_fit(double_gaussian_max, x, y, sigma=dy, p0=p0,
                                   absolute_sigma=absolute_sigma,
                                   maxfev=100000)
        cvar_b = np.diag(cvar_b)**0.5
        if popt_b[5] > popt_b[2]:
            popt_b = np.append(popt_b[3:], popt_b[:-3])
    except Exception:
        return (popt, cvar) if return_uncertainty else popt
    return (popt_b, cvar_b) if return_uncertainty else popt_b


def fit_double_gaussian_fixeddV(x, y, dy=None, return_uncertainty=None):
    """
    Same as the `fit_double_gaussian` function, but where the lines share a
    width.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.

    Returns:
        popt (array)[, cvar(array)]: The best fit parameters and, optionally,
        the uncertainty on those parameters.
    """

    # Defaults.

    dy, return_uncertainty, absolute_sigma = _errors(x, dy, return_uncertainty)
    popt = [np.nan, np.nan, np.nan, np.nan, np.nan, np.nan]
    cvar = popt.copy()

    # Initial single Gaussian fit.

    popt_a, _ = fit_gaussian(x=x, y=y, dy=dy, return_uncertainty=True)
    if not all(np.isfinite(popt_a)):
        return (popt, cvar) if return_uncertainty else popt

    # Double Gaussian fit. Try first a fit where the profile is a sum of two
    # components, then use this as strong priors for a max of two component
    # fit. This is because the max version is hard to optimize.

    p0 = [popt_a[0] + popt_a[1], popt_a[1], 0.8 * popt_a[2],
          popt_a[0] - popt_a[1], 0.8 * popt_a[2]]
    try:
        popt_b, _ = curve_fit(double_gaussian_sum_fixeddV, x, y, sigma=dy,
                              p0=p0, absolute_sigma=absolute_sigma,
                              maxfev=100000)
        p0 = [popt_b[0], popt_b[1],
              double_gaussian_sum_fixeddV(popt_b[0], *popt_b),
              popt_b[3],
              double_gaussian_sum_fixeddV(popt_b[3], *popt_b)]
        popt_b, cvar_b = curve_fit(double_gaussian_max_fixeddV, x, y, sigma=dy,
                                   p0=p0, absolute_sigma=absolute_sigma,
                                   maxfev=100000)
        cvar_b = np.diag(cvar_b)**0.5
        if popt_b[4] > popt_b[2]:
            popt_b = [popt_b[3], popt_b[1], popt_b[4], popt_b[0], popt_b[2]]
            popt_b = np.squeeze(popt_b)
    except Exception:
        return (popt, cvar) if return_uncertainty else popt
    return (popt_b, cvar_b) if return_uncertainty else popt_b


def get_gaussian_center(x, y, dy=None, return_uncertainty=None, fill=1e50):
    """
    Return the line center from a Gaussian fit to the spectrum.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.
        fill (Optional[float]): Values to return if the fit fails.

    Returns:
        popt (array)[, cvar(array)]: The best fit line center and, optionally,
        the uncertainty on the line center.
    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    popt, cvar = fit_gaussian(x, y, dy, return_uncertainty=True)
    if np.isfinite(popt[0]):
        return (popt[0], cvar[0]) if return_uncertainty else popt[0]
    return (fill, fill) if return_uncertainty else fill


def get_gaussthick_center(x, y, dy=None, return_uncertainty=None, fill=1e50):
    """
    Return the line center from an optically thick Gaussian fit to the spectrum.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.
        fill (Optional[float]): Values to return if the fit fails.

    Returns:
        popt (array)[, cvar(array)]: The best fit line center and, optionally,
        the uncertainty on the line center.
    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    popt, cvar = fit_gaussian_thick(x, y, dy, return_uncertainty=True)
    if np.isfinite(popt[0]):
        return (popt[0], cvar[0]) if return_uncertainty else popt[0]
    return (fill, fill) if return_uncertainty else fill


def get_doublegauss_center(x, y, dy=None, return_uncertainty=None, fill=1e50):
    """
    Return the line center from a fit of two Gaussians to the spectrum. The line
    center will be of the component with the largest amplitude.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.
        fill (Optional[float]): Values to return if the fit fails.

    Returns:
        popt (array)[, cvar(array)]: The best fit line center and, optionally,
        the uncertainty on the line center.
    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    popt, cvar = fit_double_gaussian(x, y, dy, return_uncertainty=True)
    if np.isfinite(popt[2]) and np.isfinite(popt[5]):
        return (popt[0], cvar[0]) if return_uncertainty else popt[0]
    else:
        return (fill, fill) if return_uncertainty else fill


def get_doublegauss_fixeddV_center(x, y, dy=None, return_uncertainty=None,
                                   fill=1e50):
    """
    Return the line center from a fit of two Gaussians which share a width to
    the spectrum. The line  center will be of the component with the largest
    amplitude.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.
        fill (Optional[float]): Values to return if the fit fails.

    Returns:
        popt (array)[, cvar(array)]: The best fit line center and, optionally,
        the uncertainty on the line center.
    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    popt, cvar = fit_double_gaussian_fixeddV(x, y, dy, return_uncertainty=True)
    if np.isfinite(popt[2]) and np.isfinite(popt[4]):
        return (popt[0], cvar[0]) if return_uncertainty else popt[0]
    else:
        return (fill, fill) if return_uncertainty else fill


def get_gaussian_width(x, y, dy=None, return_uncertainty=None, fill=1e50):
    """
    Return the absolute width of a Gaussian fit to the spectrum.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.
        dy (Optional[array]): Uncertainties on data.
        return_uncertainty (Optional[bool]): Whether to return uncertainties.
        fill (Optional[float]): Values to return if the fit fails.

    Returns:
        popt (array)[, cvar(array)]: The best fit line width and, optionally,
        the uncertainty on the line width.
    """
    if return_uncertainty is None:
        return_uncertainty = dy is not None
    popt, cvar = fit_gaussian(x, y, dy, return_uncertainty=True)
    if np.isfinite(popt[1]):
        return (popt[1], cvar[1]) if return_uncertainty else popt[1]
    return (fill, fill) if return_uncertainty else fill


def get_p0_gaussian(x, y):
    """
    Estimate (x0, dV, Tb) for the spectrum.

    Args:
        x (array): Dependent coordinate.
        y (array): Data coordinate.

    Returns:
        p0 (array): Estimate (x0, dV, Tb) values for the provided data.
    """
    if x.size != y.size:
        raise ValueError("Mismatch in array shapes.")
    Tb = np.max(y)
    x0 = x[y.argmax()]
    dV = np.trapz(y, x) / Tb / np.sqrt(2. * np.pi)
    return x0, dV, Tb


# -- MODEL FUNCTIONS --#

def gaussian(x, x0, dV, Tb):
    """Gaussian function."""
    return Tb * np.exp(-np.power((x - x0) / dV, 2.0))


def gaussian_thick(x, x0, dV, Tex, tau0):
    """Optically thick Gaussian line."""
    tau = gaussian(x, x0, dV, tau0)
    return Tex * (1. - np.exp(-tau))


def double_gaussian_sum(x, x0, dV0, Tb0, x1, dV1, Tb1):
    """Double Gaussian profile as sum of two Gaussians."""
    return gaussian(x, x0, dV0, Tb0) + gaussian(x, x1, dV1, Tb1)


def double_gaussian_max(x, x0, dV0, Tb0, x1, dV1, Tb1):
    """Double Gaussian profile as the max of two Gaussian components."""
    return np.max([gaussian(x, x0, dV0, Tb0),
                   gaussian(x, x1, dV1, Tb1)], axis=0)


def double_gaussian_sum_fixeddV(x, x0, dV, Tb0, x1, Tb1):
    """Double Gaussian profile as the sum of two Gaussians with same dV."""
    return gaussian(x, x0, dV, Tb0) + gaussian(x, x1, dV, Tb1)


def double_gaussian_max_fixeddV(x, x0, dV, Tb0, x1, Tb1):
    """Double Gaussian profile as the max of two Gaussian components."""
    return np.max([gaussian(x, x0, dV, Tb0),
                   gaussian(x, x1, dV, Tb1)], axis=0)


def SHO(x, A, C):
    """Simple harmonic oscillator."""
    return A * np.cos(x) + C


def SHO_offset(x, A, B, C):
    """Simple harmonic oscillator with offset."""
    return A * np.cos(x + B) + C


def SHO_double(x, A, B, C):
    """Two orthogonal simple harmonic oscillators."""
    return A * np.cos(x) + B * np.sin(x) + C


# -- PLOTTING FUNCTIONS -- #

def plot_walkers(samples, nburnin=None, labels=None, histogram=True):
    """
    Plot the walkers to check if they are burning in.

    Args:
        samples (ndarray):
        nburnin (Optional[int])
    """

    # Check the length of the label list.

    if labels is not None:
        if samples.shape[0] != len(labels):
            raise ValueError("Not correct number of labels.")

    # Cycle through the plots.

    for s, sample in enumerate(samples):
        fig, ax = plt.subplots()
        for walker in sample.T:
            ax.plot(walker, alpha=0.1, color='k')
        ax.set_xlabel('Steps')
        if labels is not None:
            ax.set_ylabel(labels[s])
        if nburnin is not None:
            ax.axvline(nburnin, ls=':', color='r')

        # Include the histogram.

        if histogram:
            fig.set_size_inches(1.37 * fig.get_figwidth(),
                                fig.get_figheight(), forward=True)
            ax_divider = make_axes_locatable(ax)
            bins = np.linspace(ax.get_ylim()[0], ax.get_ylim()[1], 50)
            hist, _ = np.histogram(sample[nburnin:].flatten(), bins=bins,
                                   density=True)
            bins = np.average([bins[1:], bins[:-1]], axis=0)
            ax1 = ax_divider.append_axes("right", size="35%", pad="2%")
            ax1.fill_betweenx(bins, hist, np.zeros(bins.size), step='mid',
                              color='darkgray', lw=0.0)
            ax1.set_ylim(ax.get_ylim()[0], ax.get_ylim()[1])
            ax1.set_xlim(0, ax1.get_xlim()[1])
            ax1.set_yticklabels([])
            ax1.set_xticklabels([])
            ax1.tick_params(which='both', left=0, bottom=0, top=0, right=0)
            ax1.spines['right'].set_visible(False)
            ax1.spines['bottom'].set_visible(False)
            ax1.spines['top'].set_visible(False)


def plot_corner(samples, labels=None, quantiles=[0.16, 0.5, 0.84]):
    """
    A wrapper for DFM's corner plots.

    Args:
        samples (ndarry):
        labels (Optional[list]):
        quantiles (Optional[list]): Quantiles to use for plotting.
    """
    import corner
    corner.corner(samples, labels=labels, title_fmt='.4f', bins=30,
                  quantiles=quantiles, show_titles=True)
